/*
 * Copyright (c) 2012, Willow Garage, Inc.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 *     * Redistributions of source code must retain the above copyright
 *       notice, this list of conditions and the following disclaimer.
 *     * Redistributions in binary form must reproduce the above copyright
 *       notice, this list of conditions and the following disclaimer in the
 *       documentation and/or other materials provided with the distribution.
 *     * Neither the name of the Willow Garage, Inc. nor the names of its
 *       contributors may be used to endorse or promote products derived from
 *       this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 */

#include <QColor>
#include <QSlider>
#include <QLabel>
#include <QGridLayout>
#include <QVBoxLayout>
#include<QLineEdit>
#include<QSpinBox>
#include<QPushButton>
#include <QProcess>
#include <QDebug>

#include<iostream>
#include "rviz/visualization_manager.h"
#include "rviz/render_panel.h"
#include "rviz/display.h"

#include "myviz.h"
#include "qthread.h"

// BEGIN_TUTORIAL
// Constructor for MyViz.  This does most of the work of the class.
MyViz::MyViz( QWidget* parent )
  : QWidget( parent )
{

  nh=ros::NodeHandle("~");
  sub=nh.subscribe("/test",1000,&MyViz::subCallback,this);
  pub=nh.advertise<std_msgs::String>("/test_pub",1000);
  pub_thread=std::thread(&MyViz::pubThread,this);
  //pub_thread.join();
  // Construct and lay out labels and slider controls.
  QLabel* thickness_label = new QLabel( "Line Thickness" );
  QSlider* thickness_slider = new QSlider( Qt::Horizontal );
  thickness_slider->setMinimum( 1 );
  thickness_slider->setMaximum( 100 );
  QLabel* cell_size_label = new QLabel( "Cell Size" );
  QSlider* cell_size_slider = new QSlider( Qt::Horizontal );
  cell_size_slider->setMinimum( 1 );
  cell_size_slider->setMaximum( 100 );
  QLabel* Topic=new QLabel("TOPIC:");
  QLineEdit* Topic_text=new QLineEdit();
  QGridLayout* controls_layout = new QGridLayout();
  QLabel* Size=new QLabel("Cloud Size");
  QSpinBox* size=new QSpinBox();

  /*button goto file begining*/
  QPushButton* but_reset = new QPushButton("reset", this);
  /*button goto file ending*/
  QPushButton* but_ending = new QPushButton("end", this);

  //TODO: set button color
  //button->setStyleSheet("background-color: red; color: white;");
  //TODO: set shortcut key
  //button->setShortcut(QKeySequence(Qt::CTRL + Qt::Key_A));


  //enable button
  but_reset->setEnabled(true);
  but_ending->setEnabled(true);

  but_reset->setFixedSize(100, 30);
  but_ending->setFixedSize(100, 30);

  controls_layout->addWidget( thickness_label, 0, 0 );
  controls_layout->addWidget( thickness_slider, 0, 1 );
  controls_layout->addWidget( cell_size_label, 0, 2 );
  controls_layout->addWidget( cell_size_slider, 0, 3 );

  //button locatin
  controls_layout->addWidget( but_reset, 1, 0 );
  controls_layout->addWidget( but_ending, 1, 1 );

  controls_layout->addWidget(Topic,0,4);
  controls_layout->addWidget(Topic_text,0,5);
  controls_layout->addWidget(Size,0,6);
  controls_layout->addWidget(size,0,7);


  // Construct and lay out render panel.
  render_panel_ = new rviz::RenderPanel();
  QVBoxLayout* main_layout = new QVBoxLayout;
  main_layout->addLayout( controls_layout );
  main_layout->addWidget( render_panel_ );

  // Set the top-level layout for this MyViz widget.
  setLayout( main_layout );

  // Make signal/slot connections.
  connect( thickness_slider, SIGNAL( valueChanged( int )), this, SLOT( setThickness( int )));
  connect( cell_size_slider, SIGNAL( valueChanged( int )), this, SLOT( setCellSize( int )));
  //connect(Topic_text,SIGNAL(textChanged(const QString &)),this,SLOT(setCloudTopic(const QString &)));

  //topic change signal
  connect(Topic_text,SIGNAL(textChanged(const QString &)),this,SLOT(setlaserTopic(const QString &)));
  //button signal
  connect(but_ending, &QPushButton::clicked, this,  [&]() {
    QStringList arguments, infoarg;
    QString time;
    QString file_path = "/home/noerror/123/catkin_for_ros/src/librviz_tutorial/data/result.bag";
    QProcess *rosbaginfoProcess = new QProcess(this); 

    infoarg  << "info" << file_path;
    rosbaginfoProcess->start("rosbag", infoarg);
    
    if(rosbaginfoProcess->waitForFinished()){
      QByteArray output = rosbaginfoProcess->readAllStandardOutput();
      QString outputString = QString::fromLocal8Bit(output);
      QString delimiter = "duration:";
      outputString = outputString.section(delimiter, 1, 1);
      delimiter = "\n";
      outputString = outputString.section(delimiter, 0, 0);
      outputString.remove(" ");
      outputString.remove("s");
      double number = outputString.toDouble(); //qstring convert to double
      time = QString::number(number-0.1);
      qDebug() << time;
    }
    
    arguments << "play" << file_path << "--start" << time;
    QProcess *rosbagProcess = new QProcess(this); 
    //QProcess::execute("rosbag play /home/noerror/123/catkin_for_ros/src/librviz_tutorial/data/result.bag");

    rosbagProcess->start("rosbag", arguments);
    
    });
  connect(but_reset, &QPushButton::clicked, this,  [&]() {
    QStringList arguments, infoarg;
    QString time;
    QString file_path = "/home/noerror/123/catkin_for_ros/src/librviz_tutorial/data/result.bag";
    
    arguments << "play" << file_path;
    QProcess *rosbagProcess = new QProcess(this); 
    //QProcess::execute("rosbag play /home/noerror/123/catkin_for_ros/src/librviz_tutorial/data/result.bag");

    rosbagProcess->start("rosbag", arguments);
    
    });
  connect(size,SIGNAL(valueChanged(int)),this,SLOT(setCloudSize(int)));
  // Next we initialize the main RViz classes.
  //
  // The VisualizationManager is the container for Display objects,
  // holds the main Ogre scene, holds the ViewController, etc.  It is
  // very central and we will probably need one in every usage of
  // librviz.
  manager_ = new rviz::VisualizationManager( render_panel_ );
  render_panel_->initialize( manager_->getSceneManager(), manager_ );
  //manager_->setFixedFrame("world");

  /*TODO: laser data*/
  manager_->setFixedFrame("laser");
  manager_->initialize();
  manager_->startUpdate();

  // Create a Grid display.
  grid_ = manager_->createDisplay( "rviz/Grid", "adjustable grid", true );
  ROS_ASSERT( grid_ != NULL );
  
  cloud = manager_->createDisplay( "rviz/PointCloud2", "point cloud", true );
  ROS_ASSERT( cloud != NULL );

  /*TODO: laser data display*/
  laser = manager_->createDisplay( "rviz/LaserScan", "laser", true );
  ROS_ASSERT( laser != NULL );
  //set defualt topic to /scan
  laser->subProp("Topic")->setValue("/scan");
  laser->subProp("Decay Time")->setValue("0");


  cloud->subProp("Topic")->setValue("/pose_graph/octree");
  cloud->subProp("Style")->setValue("Points");
  cloud->subProp("Size (Pixels)")->setValue("2");
  cloud->subProp("Color Transformer")->setValue("Intensity");
  cloud->subProp("Invert Rainbow")->setValue("true");
  cloud->subProp("Decay Time")->setValue("1");
  // Configure the GridDisplay the way we like it.
  grid_->subProp( "Line Style" )->setValue( "Billboards" );
  grid_->subProp( "Color" )->setValue( QColor( Qt::yellow ) );

  // Initialize the slider values.
  thickness_slider->setValue( 5 );
  cell_size_slider->setValue( 10 );
}

// Destructor.
MyViz::~MyViz()
{
  delete manager_;
}

// This function is a Qt slot connected to a QSlider's valueChanged()
// signal.  It sets the line thickness of the grid by changing the
// grid's "Line Width" property.
void MyViz::setThickness( int thickness_percent )
{
  if( grid_ != NULL )
  {
    grid_->subProp( "Line Style" )->subProp( "Line Width" )->setValue( thickness_percent / 100.0f );
  }
}

// This function is a Qt slot connected to a QSlider's valueChanged()
// signal.  It sets the cell size of the grid by changing the grid's
// "Cell Size" Property.
void MyViz::setCellSize( int cell_size_percent )
{
  if( grid_ != NULL )
  {
    grid_->subProp( "Cell Size" )->setValue( cell_size_percent / 10.0f );
  }
}
void MyViz::setCloudTopic(const QString &newTopic){
  if(cloud!=NULL){
    cloud->subProp( "Topic" )->setValue(newTopic);
    //ROS_INFO_STREAM("cloud topic changed to => "<<newTopic.toStdString());
  }
}
void MyViz::setCloudSize(int cloudsize){
  if(cloud!=NULL){
    cloud->subProp("Size (Pixels)")->setValue(cloudsize);
  }
}
void MyViz::subCallback(const std_msgs::String& msg){
  ROS_INFO_STREAM("receive message!");
}
void MyViz::pubThread(){
  while(ros::ok()){
    ROS_INFO_STREAM_ONCE("here is in publish process!");
  }
}
void MyViz::setlaserTopic(const QString & lasertopic){
  if(laser!=NULL){
    laser->subProp("Topic")->setValue(lasertopic);
    manager_->startUpdate();
  }
    
}
void MyViz::PushButtonBegining(const QString & things){
  
  QProcess::execute("rosbag play /home/noerror/123/catkin_for_ros/src/librviz_tutorial/data/result.bag ");
  
}